//+------------------------------------------------------------------+
//|                                                       canvas.mqh |
//|                        Copyright 2012, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, MetaQuotes Software Corp."
#property link      "http://www.mql5.com"
//+------------------------------------------------------------------+
//|                                          |
//+------------------------------------------------------------------+
#define XRGB(r,g,b)    (0xFF000000|(uchar(r)<<16)|(uchar(g)<<8)|uchar(b))
#define ARGB(a,r,g,b)  ((uchar(a)<<24)|(uchar(r)<<16)|(uchar(g)<<8)|uchar(b))
#define ARGBN(a,r,g,b) ((uchar(a)<<24)|(uchar(r*a/255)<<16)|(uchar(g*a/255)<<8)|uchar(b*a/255))
#define GETRGBA(col)   uchar((col)>>24)
#define GETRGBR(col)   uchar((col)>>16)
#define GETRGBG(col)   uchar((col)>>8)
#define GETRGBB(col)   uchar(col)

uint AMixColor(uint bk, uint col)
{
  uint bkA= ARGBN(GETRGBA(bk),GETRGBR(bk),GETRGBG(bk),GETRGBB(bk));
  uint  alpha = col>>24;
  uchar r=uchar(((uint)GETRGBR(bkA)*(255-alpha)+(uint)GETRGBR(col)*alpha)/255);
  uchar g=uchar(((uint)GETRGBG(bkA)*(255-alpha)+(uint)GETRGBG(col)*alpha)/255);
  uchar b=uchar(((uint)GETRGBB(bkA)*(255-alpha)+(uint)GETRGBB(col)*alpha)/255);
  return ARGB(alpha,r,g,b);
}
//+------------------------------------------------------------------+
//|                               |
//+------------------------------------------------------------------+
class CCanvas
  {
   string            m_name;
   string            m_rcname;
   int               m_width;
   int               m_height;
   ENUM_COLOR_FORMAT m_format;
   uint              m_data[];

public:
                     CCanvas();
                    ~CCanvas() { Destroy(); }
   //--- /
   bool              Create(string name,int x,int y,int width,int height,ENUM_COLOR_FORMAT clrfmt);
   void              Destroy();
   //---    
   void              Update();
   //--- / 
   void              Erase(uint col=0) { ArrayInitialize(m_data,col); }
   //---  
   uint              GetPixel(int x,int y) const;
   void              SetPixel(int x,int y,uint col);
   void              Line(int x1,int y1,int x2,int y2,uint col);
   void              Rect(int x1,int y1,int x2,int y2,uint col);
   void              Circle(int x,int y,int r,uint col);
   void              Triangle(int x1,int y1,int x2,int y2,int x3,int y3,uint col);
   //---   
   void              FillRect(int x1,int y1,int x2,int y2,uint col);
   void              FillCircle(int x,int y,int r,uint col);
   void              FillTriangle(int x1,int y1,int x2,int y2,int x3,int y3,uint col);
  };
//+------------------------------------------------------------------+
//|                                                       |
//+------------------------------------------------------------------+
CCanvas::CCanvas():m_name(NULL),m_rcname(NULL),m_width(0),m_height(0),m_format(COLOR_FORMAT_XRGB_NOALPHA)
  {
  }
//+------------------------------------------------------------------+
//|              |
//+------------------------------------------------------------------+
bool CCanvas::Create(string name,int x,int y,int width,int height,ENUM_COLOR_FORMAT clrfmt)
  {
   Destroy();
//---   
   if(width>0 && height>0 && ArrayResize(m_data,width*height)>0)
     {
      //---   
      if(ObjectCreate(0,name,OBJ_BITMAP_LABEL,0,0,0))
        {
         //---   
         m_rcname="::"+(string)GetTickCount();
         //---   
         ArrayInitialize(m_data,0);
         //---   
         if(ResourceCreate(m_rcname,m_data,width,height,0,0,0,clrfmt))
           {
            //---  x,y     
            if(ObjectSetInteger(0,name,OBJPROP_XDISTANCE,x) && 
               ObjectSetInteger(0,name,OBJPROP_YDISTANCE,y) && 
               ObjectSetString(0,name,OBJPROP_BMPFILE,m_rcname))
              {
               //---   
               //---  
               m_width =width;
               m_height=height;
               m_name  =name;
               m_format=clrfmt;
               //---
               return(true);
              }
           }
        }
     }
//---  -   
   Destroy();
   return(false);
  }
//+------------------------------------------------------------------+
//|                   |
//+------------------------------------------------------------------+
void CCanvas::Destroy(void)
  {
//---   
   ObjectDelete(0,m_name);
//---  
   ArrayFree(m_data);
//---  
   m_width  =0;
   m_height =0;
   m_name   =NULL;
   m_rcname =NULL;
  }
//+------------------------------------------------------------------+
//|     ()                       |
//+------------------------------------------------------------------+
void CCanvas::Update(void)
  {
//---     
   if(ResourceCreate(m_rcname,m_data,m_width,m_height,0,0,0,m_format))
      ChartRedraw();
  }
//+------------------------------------------------------------------+
//|                                                    |
//+------------------------------------------------------------------+
void CCanvas::SetPixel(int x,int y,uint col)
  {
//---  
   if(x>=0 && x<m_width && y>=0 && y<m_height)
      m_data[y*m_width+x]=AMixColor(m_data[y*m_width+x],col);
  }
//+------------------------------------------------------------------+
//|                                               |
//+------------------------------------------------------------------+
uint CCanvas::GetPixel(int x,int y) const
  {
//---  
   if(x>=0 && x<m_width && y>=0 && y<m_height)
      return(m_data[y*m_width+x]);
//--- 
   return(0);
  }
//+------------------------------------------------------------------+
//|                                |
//+------------------------------------------------------------------+
void CCanvas::Line(int x1,int y1,int x2,int y2,uint col)
  {
//---    
   if((x1<0 && x2<0) || (y1<0 && y2<0)) return;
   if(x1>=m_width && x2>=m_width)       return;
   if(y1>=m_height && y2>=m_height)     return;
//---  y  x  y
   int dx=(int)MathAbs(x2-x1);
   int dy=(int)MathAbs(y2-y1);
//---    x  y
   int sx=x1<x2?1:-1;
   int sy=y1<y2?1:-1;
   int er=dx-dy;
//---   
   while(x1!=x2 || y1!=y2)
     {
      SetPixel(x1,y1,col);
      //---    
      int er2=er<<1;
      if(er2>-dy) { er-=dy; x1+=sx; }
      if(er2< dx) { er+=dx; y1+=sy; }
     }
//---    
   SetPixel(x2,y2,col);
  }
//+------------------------------------------------------------------+
//|                                           |
//+------------------------------------------------------------------+
void CCanvas::Rect(int x1,int y1,int x2,int y2,uint col)
  {
   int x,y;
//---  
   if(x2<x1) { x=x1; x1=x2; x2=x; }
   if(y2<y1) { y=y1; y1=y2; y2=y; }
//---    
   if(x2>=0 && x1<m_width)
     {
      //---   
      if(y1>=0 && y1<m_height)
         for(x=(int)MathMax(x1,0),y=MathMin(x2,m_width-1);x<y;x++)
            m_data[y1*m_width+x]=col;
      //---   
      if(y2>=0 && y2<m_height)
         for(x=(int)MathMax(x1,0),y=MathMin(x2,m_width-1);x<y;x++)
            m_data[y2*m_width+x]=col;
     }
//---    
   if(y2>=0 && y1<m_height)
     {
      //---   
      if(x1>=0 && x1<m_width)
         for(y=(int)MathMax(y1,0),x=MathMin(y2,m_height-1);y<x;y++)
            m_data[y*m_width+x1]=col;
      //---   
      if(x2>=0 && x2<m_width)
         for(y=(int)MathMax(y1,0),x=MathMin(y2,m_height-1);y<x;y++)
            m_data[y*m_width+x2]=col;
     }
  }
//+------------------------------------------------------------------+
//|                           |
//+------------------------------------------------------------------+
void CCanvas::Circle(int x,int y,int r,uint col)
  {
   int dx=0;
   int dy=r;
   int dd=2-2*r;
   int er=0;
//---
   while(dy>=0)
     {
      SetPixel(x+dx,y+dy,col);
      SetPixel(x+dx,y-dy,col);
      SetPixel(x-dx,y+dy,col);
      SetPixel(x-dx,y-dy,col);
      //---
      er=2*(dd+dy)-1;
      if(dd<0 && er<=0)
        {
         dx++;
         dd+=2*dx+1;
         continue;
        }
      //---
      er=2*(dd-dx)-1;
      if(dd>0 && er>0)
        {
         dy--;
         dd+=1-2*dy;
         continue;
        }
      dx++;
      dd+=2*(dx-dy);
      dy--;
     }
  }
//+------------------------------------------------------------------+
//|                                             |
//+------------------------------------------------------------------+
void CCanvas::Triangle(int x1,int y1,int x2,int y2,int x3,int y3,uint col)
  {
   Line(x1,y1,x2,y2,col);
   Line(x2,y2,x3,y3,col);
   Line(x3,y3,x1,y1,col);
  }
//+------------------------------------------------------------------+
//|                                     |
//+------------------------------------------------------------------+
void CCanvas::FillCircle(int x,int y,int r,uint col)
  {
   int dx=0;
   int dy=r;
   int dd=2-2*r;
   int er=0;
   int x1,x2,y1,y2,t;
//---
   while(dy>=0)
     {
      x1=x-dx;
      x2=x+dx;
      y1=y+dy;
      y2=y-dy;
      //---    
      if(x1>x2) { t=x1; x1=x2; x2=t; }
      if(x1<0) x1=0;
      if(x2>=m_width) x2=m_width-1;
      //---   
      if(y1>=0 && y1<m_height)
         for(int x0=x1;x0<x2;x0++)
            m_data[y1*m_width+x0]=col;
      //---   
      if(y2>=0 && y2<m_height)
         for(int x0=x1;x0<x2;x0++)
            m_data[y2*m_width+x0]=col;
      //---
      er=2*(dd+dy)-1;
      if(dd<0 && er<=0)
        {
         dx++;
         dd+=2*dx+1;
         continue;
        }
      //---
      er=2*(dd-dx)-1;
      if(dd>0 && er>0)
        {
         dy--;
         dd+=1-2*dy;
         continue;
        }
      dx++;
      dd+=2*(dx-dy);
      dy--;
     }
  }
//+------------------------------------------------------------------+
//|                               |
//+------------------------------------------------------------------+
void CCanvas::FillRect(int x1,int y1,int x2,int y2,uint col)
  {
   int x;
//---  
   if(x2<x1) { x=x1; x1=x2; x2=x; }
   if(y2<y1) { x=y1; y1=y2; y2=x; }
//---    
   if(x1<0) x1=0;
   if(y1<0) y1=0;
   if(x2>=m_width ) x2=m_width -1;
   if(y2>=m_height) y2=m_height-1;
//---
   for(;y1<y2;y1++)
     {
      for(x=x1;x<x2;x++)
        {
         m_data[y1*m_width+x]= AMixColor(m_data[y1*m_width+x], col);
        }
     }
  }
//+------------------------------------------------------------------+
//|                                             |
//+------------------------------------------------------------------+
void CCanvas::FillTriangle(int x1,int y1,int x2,int y2,int x3,int y3,uint col)
  {
   int    xx1,xx2,t;
   double k1=0,k2=0,xd1,xd2;
//---      
   if(y1>y2) { t=y2; y2=y1; y1=t; t=x2; x2=x1; x1=t; }
   if(y1>y3) { t=y1; y1=y3; y3=t; t=x1; x1=x3; x3=t; }
   if(y2>y3) { t=y2; y2=y3; y3=t; t=x2; x2=x3; x3=t; }
//---     
   if(y3<0 ||  y1>m_height) return;
   if(x1<0 && x2<0 && x3<0) return;
   if(x1>m_width && x2>m_width && x3>m_width) return;
//---   
   if((t=y1-y2)!=0) k1=(x1-x2)/(double)t;
   if((t=y1-y3)!=0) k2=(x1-x3)/(double)t;
//---
   xd1=x1;
   xd2=x1;
//---
   for(int i=y1;i<=y3;i++)
     {
      if(i==y2)
        {
         if((t=y2-y3)!=0) k1=(x2-x3)/(double)t;
         xd1=x2;
        }
      //---     
      xx1=(int)xd1; xd1+=k1;
      xx2=(int)xd2; xd2+=k2;
      //---    
      if(i<0 || i>=m_height) continue;
      //--- 
      if(xx1>xx2)
        {
         t=xx1;
         xx1=xx2;
         xx2=t;
        }
      //---    
      if(xx2<0 || xx1>=m_width) continue;
      //---   ,    
      if(xx1<0) xx1=0;
      if(xx2>=m_width) xx2=m_width-1;
      //---    
      for(int j=xx1;j<=xx2;j++)
        {
         m_data[i*m_width+j]=col;
        }
     }
  }
//+------------------------------------------------------------------+//+------------------------------------------------------------------+
